(* solve the Minimum d-Bounded-Degree Deletion problem 
   Copyright (C) 2009 Hannes Moser, Manuel Sorge

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(** Parameters that may be set from outside. *)
val nt_kernelization : bool ref
val exhaustive_interleaving : bool ref
val buss_interleaving : bool ref
val deg_one_interleaving : bool ref
val deg_two_interleaving : bool ref
val degree_less_than_or_equal_d_interleaving : bool ref
val cliquerule_interleaving : bool ref
val guided_branch_vertex_selection : bool ref
val greedysol_branching : bool ref
val forbid_branch_vertices : bool ref
val edge_rule : bool ref
val bound_rule : bool ref
val improve_lower_bound : bool ref
val forbid_sol_subsets : bool ref
val min_branch_candidates : int ref
val interleaving_interval : int ref
val greedyopt_interval : int ref

val lb_executions : int ref
val lb_improvements : int ref

(** [lower_bound_f g d] returns a greedily computed lower bound for the
    d-Bounded-Degree Deletion instance g. *)
val lower_bound_f : FGraph.t -> int -> int

(** [bddsolve_greedy_f g d] returns a greedily computed solution for the
    d-Bounded-Degree Deletion instance g. (This represents an upper bound for
    the minimum solution.) *)
val bddsolve_greedy_f : FGraph.t -> int -> IntSet.t

(** [bddkernel_buss g d k] kernelizes the d-Bounded-Degree Deletion instance 
    (g, k) (g graph and k solution size parameter) with the high-degree rule.
    It returns (s, g', k') where s contains the new solution vertices and
    (g', k') is the reduced instance. *)
val bddkernel_buss : FGraph.t -> int -> int -> IntSet.t * FGraph.t * int

(** Variables to keep track of the reduction rule efficiency. *) 
val buss_num : int ref
val buss_vertices_deleted : int ref
val deg_one_num : int ref
val deg_one_vertices_deleted : int ref
val deg_two_num : int ref
val cliquerule_num : int ref
val cliquerule_vertices_deleted : int ref
val ntrule_num : int ref
val ntrule_vertices_deleted : int ref

(** Variables counting search tree nodes and those that were leafes due to 
    various rules and other reasons.*)
val nodecount : int ref
val kernelnodecount : int ref
val bounded_degree_nodecount : int ref
val finished_after_simple_interleaving_nodecount : int ref
val no_instance_after_simple_interleaving_nodecount : int ref
val bound_rule_nodecount : int ref
val edge_rule_nodecount : int ref
val forbidden_nodecount : int ref
val no_instance_due_forbidden_vertices_nodecount : int ref
val no_instance_due_branching_nodecount : int ref

(** Variables accounting for time consumption. *)
val rrules_time : float ref
val ntrule_time : float ref
val branch_vertex_selection_time : float ref
val greedy_sol_time : float ref
val preprocess_time : float ref
val edge_rule_time : float ref
val lb_rule_time : float ref

(** Enumerate all solutions - currently not used. *)
(* val bddsolve : FGraph.t -> int -> int -> IntSet.t *)

(** [bddsolve' g d k time_limit] returns a solution for the d-Bounded-Degree Deletion
    instance (g, k) or raises NO_instance if there is none. If the computation takes 
    longer than time_limit seconds, Time_Out is raised. *)
val bddsolve' : FGraph.t -> int -> int -> float -> IntSet.t

(** [bddsolve_optimal g d time_limit] returns a optimal solution for the 
    d-Bounded-Degree Deletion instance g. If the computation takes longer than 
    time_limit seconds, an upper and lower bound for the optimal solution is returned.
    *)
val bddsolve_optimal : FGraph.t -> int -> float -> IntSet.t * int * int


(** Exception raised, if the given instance has no solution with size <= k. *)
exception NO_instance

(** Exception raised, if the given time limit was violated. *)
exception Time_out
