(* F(ancy)Graph -- A wrapper for Graph.ml by Falk Hueffner
   Copyright (C) 2009 Manuel Sorge

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

let xor a b = (a && (not b)) || ((not a) && b);;

(* Main *)

type extras = {
  vset: IntSet.t; (* The IntSet of vertices present in the graph. *)
  comp: bool; (* Shall functions return the values of the complement to the graph actually saved? *)
  edge_n: int; (* Number of edges present in the saved graph. *)
};;

type t = Graph.t * extras;;

let empty = 
  (Graph.empty, {vset = IntSet.empty; comp = false; edge_n = 0;})
;;

let of_graph g = 
  (g, {vset = Graph.vertices g; comp = false; edge_n = Graph.num_edges g})
;;

let to_graph (g, e) = 
  if e.comp then Graph.complement g
  else g
;;

let connector g i s =
  IntSet.fold begin fun g j -> Graph.connect g i j end s g
;;

(* 'Translate' degree, edge number and neighbors between the saved and the represented graph.*)
let deg_trans e d = if e.comp then ((IntSet.size e.vset) - d) - 1 else d;;
let deg_trans' e d = ((IntSet.size e.vset) - d) - 1;;

let edg_trans e m = if e.comp then let n = IntSet.size e.vset in ((n * (n - 1)) / 2) - m else m;;
let edg_trans' e m = let n = IntSet.size e.vset in ((n * (n - 1)) / 2) - m;;

let n_trans e i n = if e.comp then IntSet.delete (IntSet.minus e.vset n) i else n;;
let n_trans' e i n = IntSet.delete (IntSet.minus e.vset n) i;;


let add_vertex (g, e) v = 
  let g = Graph.add_vertex g v in
  let g, edge_n = if e.comp then connector g v e.vset, e.edge_n + IntSet.size e.vset else g, e.edge_n in
  let vset = IntSet.add e.vset v in
    (g, {e with vset = vset; edge_n = edge_n})
;;

(* TODO: let disconnect *)

let has_vertex (g, e) = Graph.has_vertex g;;

let connect (g, e) i j = 
  let g, edge_n = 
    if e.comp then 
      if Graph.has_vertex g i && Graph.has_vertex g j then
	try Graph.disconnect g i j, e.edge_n - 1 
	with Not_found -> raise IntSet.Already_present
      else raise Not_found
    else Graph.connect g i j, e.edge_n + 1 
  in
    (g, {e with edge_n = edge_n})
;;

let is_connected (g, e) i j =
  xor (e.comp) (Graph.is_connected g i j)
;;

let delete_vertex (g, e) i =
  let n = Graph.neighbors g i in
  let g = Graph.delete_vertex g i in
  let vset = IntSet.delete e.vset i in
  let edge_n = e.edge_n - (IntSet.size n) in
    (g, {e with edge_n = edge_n; vset = vset})
;;
    
let delete_vertices (g, e) s =
   IntSet.fold (fun accu i -> delete_vertex accu i) s (g, e)
;;

let deg_vertices (g, e) d = 
  let d = deg_trans e d in
  let d_v =
    Graph.fold_vertices begin fun d_v i n ->
      if IntSet.size n = d then IntSet.add d_v i
      else d_v
    end g IntSet.empty
  in
    d_v
;;

let deg_rel (g, e) rel = 
  let rel = if e.comp then fun d -> rel (deg_trans' e d) else rel in
  let d_v =
    Graph.fold_vertices begin fun d_v i n ->
      if rel (IntSet.size n) then IntSet.add d_v i
      else d_v
    end g IntSet.empty
  in
    d_v
;;

let deg_struct (g, e) = 
  let deg = if e.comp then fun d -> (deg_trans' e d) else fun d -> d in
  Graph.fold_vertices begin fun ds i n ->
    IntMap.modify_default begin fun d_v -> IntSet.add d_v i end ds (deg (IntSet.size n))IntSet.empty
  end g IntMap.empty
;;

let max_deg (g, e) =
  let max_deg =
    if g = Graph.empty then 0 
    else
      let compare = if e.comp then (<) else (>) in
      let m, m_n =
	Graph.fold_vertices begin fun (m, m_n) i n ->
	  if compare (IntSet.size n) (IntSet.size m_n) then i, n
	  else m, m_n
	end g (let v = Graph.max_vertex g in (v, Graph.neighbors g v))
      in
	IntSet.size m_n
  in
    deg_trans e max_deg
;;

let max_deg_vertices (g, e) = 
  let compare = if e.comp then (<) else (>) in
  let d, d_v =
    Graph.fold_vertices begin fun (d, d_v) i n ->
      if compare (IntSet.size n) d then (IntSet.size n, IntSet.singleton i)
      else if IntSet.size n = d then (d, IntSet.add d_v i)
      else d, d_v
    end g (0, IntSet.empty)
  in
    d_v
;;

let max_deg_vertex (g, e) = 
  let compare = if e.comp then (<) else (>) in
  let m, m_n =
    Graph.fold_vertices begin fun (m, m_n) i n ->
      if compare (IntSet.size n) (IntSet.size m_n) then i, n
      else m, m_n
    end g (Graph.max_vertex g, Graph.neighbors g (Graph.max_vertex g))
  in
    m
;;

(* let fold_degrees f (_, es) a = deg_fold f es.ds a;; *)



let num_edges (g, e) = edg_trans e e.edge_n;;

let deg (g, e) i = deg_trans e (Graph.deg g i);;

let neighbors (g, e) i = n_trans e i (Graph.neighbors g i);;

let neighbors' (g, e) i = 
  if e.comp then IntSet.minus e.vset (Graph.neighbors g i) 
  else Graph.neighbors' g i
;;

let fold_vertices f (g, e) s = 
  (* The neighbor-translation produces overhead, if it is not used.
     It is recommended to use IntSet.fold f (FGraph.vertices g) s instead, 
     if the neighbors are not needed. *)
  let f = if e.comp then fun accu i n -> f accu i (n_trans' e i n) else f in
    Graph.fold_vertices f g s
;;

let fold_edges f (g, e) s = 
  (* Slow!? *)
  if e.comp then 
    Graph.fold_vertices begin fun accu i _ ->
      Graph.fold_vertices begin fun accu j _ ->
	if i <= j && is_connected (g, e) i j then f accu i j
	else accu
      end g accu
    end g s
  else Graph.fold_edges f g s
;;

let fold_neighbors f (g, e) i s =
  let n = neighbors (g, e) i in
    IntSet.fold f n s
;;

let vertices (g, e) = e.vset;;
let vertex_set = vertices;;

let new_vertex (g, e) =
  let g, i = Graph.new_vertex g in
  let g = if e.comp then connector g i e.vset else g in
  let edge_n = if e.comp then e.edge_n + IntSet.size e.vset else e.edge_n in
  let vset = IntSet.add e.vset i in
    (g, {e with vset = vset; edge_n = edge_n}), i
;;

let max_vertex (g, e) = Graph.max_vertex g;;

let num_vertices (g, e) = IntSet.size e.vset;;

let subgraph (g, e) vs = 
  let sg = Graph.subgraph g vs in
    sg, {vset = vs; comp = e.comp; edge_n = Graph.num_edges sg}
;;

let complement_fake (g, e) = (g, {e with comp = not e.comp});;

let complement_real (g, e) =
  (Graph.complement g, {e with edge_n = edg_trans' e e.edge_n})
;;

let complement ((g, e) as f) =
  let f_n = float_of_int (num_vertices f) in
  let f_m = float_of_int (e.edge_n) in
  let density = f_m /. ((f_n *. (f_n -. 1.0)) /. 2.0) in
    if density > 0.5 then complement_real f else complement_fake f
;;
