(* Copyright (C) 2006  Falk Hffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

type t = IntSet.t IntMap.t;;

let empty = IntMap.empty;;

let has_vertex = IntMap.has_key;;
let num_vertices = IntMap.size;;
let vertices = IntMap.keys;;
let neighbors = IntMap.get;;
let neighbors' g v = IntSet.add (neighbors g v) v;;
let is_deg0 g i = IntSet.is_empty (neighbors g i);;
let deg g i = IntSet.size (neighbors g i);;
let max_vertex = IntMap.max_key;;

let add_vertex g i = IntMap.add g i IntSet.empty;;
let new_vertex g =
  let i = if IntMap.is_empty g then 0 else (IntMap.max_key g) + 1 in
    add_vertex g i, i
;;

let delete_vertex g i =
  let n = neighbors g i in
  let g = IntMap.delete g i in
    IntSet.fold
      (fun g j ->
	 IntMap.modify (fun n -> IntSet.delete n i) g j)
      n
      g      
;;

let is_connected g i j = IntSet.contains (IntMap.get g i) j;;

let choose_edge g =
  let i, neighbors =
    IntMap.find (fun _ neighbors -> not (IntSet.is_empty neighbors)) g
  in
    i, IntSet.choose neighbors
;;

let choose_edge_opt g =
  IntMap.find_opt
    (fun i neighbors ->
       if IntSet.is_empty neighbors then None
       else Some (i, IntSet.choose neighbors))
    g
;;

let connect g i j =
  let neighbors_i = neighbors g i in
  let neighbors_j = neighbors g j in
  let neighbors_i' = IntSet.add neighbors_i j in
  let neighbors_j' = IntSet.add neighbors_j i in
  let g = IntMap.set g i neighbors_i' in
  let g = IntMap.set g j neighbors_j' in
    g
;;

let connect' g i j =
  let g = IntMap.modify_default (fun neighbors -> IntSet.add neighbors j) g i IntSet.empty in
  let g = IntMap.modify_default (fun neighbors -> IntSet.add neighbors i) g j IntSet.empty in
    g
;;

let disconnect g i j =
  let neighbors_i = neighbors g i in
  let neighbors_j = neighbors g j in
  let neighbors_i' = IntSet.delete neighbors_i j in
  let neighbors_j' = IntSet.delete neighbors_j i in
  let g = IntMap.set g i neighbors_i' in
  let g = IntMap.set g j neighbors_j' in
    g
;;

let fold_vertices = IntMap.fold;;
let iter_vertices f g = fold_vertices (fun () i neighbors -> f i neighbors) g ();;

let vertex_set g = fold_vertices (fun s i _ -> IntSet.add s i) g IntSet.empty;;

let fold_neighbors f g i = IntSet.fold f (neighbors g i);;
let iter_neighbors f g i = IntSet.iter f (neighbors g i);;

let fold_edges f g accu =
  fold_vertices
    (fun accu i neighbors ->
       IntSet.fold
         (fun accu j -> if i <= j then f accu i j else accu)
         neighbors
         accu)
    g
    accu
;;

let iter_edges f g = fold_edges (fun () i j -> f i j) g ();;

let num_edges g =
  let num =
    fold_vertices
      (fun num v neighbors -> num + IntSet.size neighbors + if is_connected g v v then 1 else 0) g 0
  in
    assert (num mod 2 = 0);
    num / 2
;;

let complement g =
  let vertices = vertices g
  in
    IntMap.map (fun i neighbors ->
		  IntSet.delete (IntSet.minus vertices neighbors) i) g
;;

let subgraph g vs =
  IntSet.fold
    (fun g' v ->
       IntMap.add g' v (IntSet.intersection (neighbors g v) vs))
    vs
    empty
;;

let output channel g =
  Printf.fprintf channel "# n = %d, m = %d\n" (num_vertices g) (num_edges g);
  (* List degree-0 vertices.  *)
  iter_vertices (fun i neighbors ->
                   if IntSet.is_empty neighbors
                   then Printf.fprintf channel "%d\n" i) g;
  iter_edges (fun i j -> Printf.fprintf channel "%d %d\n" i j) g;
(*  Printf.fprintf channel "}\n"; *)
;;

type parsed = Nothing | Vertex of string | Edge of string * string;;

let line_parser_fsu line lineno =
  let res =
    if line.[0] = '#' then Nothing
    else
      let words = Util.split_string line in
	match List.length words with
	    1 -> Vertex (List.hd words)
	  | 2 -> Edge ((List.hd words), (List.nth words 1))
	  | _ -> (Printf.eprintf "warning: ignoring malformed line %d\n%!" lineno; Nothing)
  in
    [res]
;;

let line_parser_dimacs_asc line lineno =
  let res =
    match line.[0] with
	'n' -> Vertex (List.nth (Util.split_string line) 1)
      | 'e' -> 
	  let words = Util.split_string line in
	    Edge (List.nth words 1, List.nth words 2)
      | _ -> Nothing
	  (* None of the optional descriptors are currently supported. *)
  in
    [res]
;;


module StringMap = Map.Make(String);;

let input channel line_parser =
  let g = ref empty in
  let vertex_numbers = Hashtbl.create 100 in
  let n = ref 0 in
  let rec loop lineno =
    let line = input_line channel in
    let parsed = line_parser line lineno in
      
    let vertex_number s =
      try
	Hashtbl.find vertex_numbers s
      with Not_found ->
	Hashtbl.replace vertex_numbers s !n;
	incr n;
	!n - 1
      in
	
	List.iter 
	(function
	     | Nothing -> ()
	     | Vertex s -> 
		 (try g := add_vertex !g (vertex_number s)
		  with IntMap.Already_present -> 
		    Printf.eprintf "warning: additional vertex declaration in line %d\n%!" lineno)
	     | Edge (s1, s2) -> 
		 (let v1 = vertex_number s1
		  and v2 = vertex_number s2 in
		    if v1 = v2 then
		      Printf.eprintf "warning: ignoring self loop in line %d\n%!" lineno
		    else begin
		      try
			g := connect' !g v1 v2;
		      with IntSet.Already_present ->
			Printf.eprintf "warning: multiple edge in line %d\n%!" lineno
		    end))
	  parsed;
	
	loop (lineno + 1)
    in
      try
	loop 1
      with End_of_file ->
	let vertex_names = Array.make !n "" in
	  Hashtbl.iter (fun s v -> vertex_names.(v) <- s) vertex_numbers;
	  (*	output stdout !g; *)
	  !g, vertex_numbers, vertex_names
;;

let input_n channel =
  let g = ref empty in
  let rec loop lineno =
    let line = input_line channel in
      if line.[0] = '#' then loop (lineno + 1)
      else
	let p = let rec loop p =
	  if Util.is_whitespace line.[p] then p else loop (p + 1) in loop 0 in
	let q = let rec loop q =
	  if not (Util.is_whitespace line.[q]) then q else loop (q + 1) in loop (p + 1) in
	let s1 = String.sub line 0 p in
	let s2 = String.sub line q (String.length line - q) in
	let v1 = int_of_string s1 in
	let v2 = int_of_string s2 in
	  if v1 = v2 then
	     Printf.eprintf "warning: ignoring self loop in line %d\n%!" lineno
	  else begin
	    try
	       g := connect' !g v1 v2;
	     with IntSet.Already_present ->
	       Printf.eprintf "warning: multiple edge in line %d\n%!" lineno
	  end;
	  loop (lineno + 1)
  in
    try
      loop 1
    with End_of_file -> !g
;;

let input_slow channel =
  let strip_comment s =
    if String.contains s '#' then String.sub s 0 (String.index s '#') else s in
  let isdigit = function
      '0' .. '9' -> true
    | _ -> false in
  (* try int_of_string... is not what we want, since it accepts e.g. "0x17"  *)
  let is_int s =
    let rec loop p =
      if p >= String.length s
      then true
      else isdigit s.[p] && loop (p + 1)
    in
      String.length s > 0 && loop 0 in
  let rec loop lines lineno =
    try
      let line = input_line channel in
      let line = strip_comment line in
      let line = Util.split_string line in
	if line = [] then loop lines (lineno + 1)
	else
	  let v, w =
	    match line with
	      | [v; w]    -> v, w
	      | _ -> invalid_arg "bad edge syntax"
	  in
	    loop ((v, w) :: lines) (lineno + 1)
    with End_of_file -> lines in
  let lines = loop [] 0 in
  let vertex_numbers, vertex_names, max_id =
    List.fold_left
      (fun (vertex_numbers, vertex_names, max_id) (v, w) ->
	 let update vertex_numbers vertex_names max_id v =
	   if is_int v && not (StringMap.mem v vertex_numbers)
	   then
	     let i = int_of_string v
	     in
	       StringMap.add v i vertex_numbers,
	       IntMap.add vertex_names i v,
	       max max_id i
	   else vertex_numbers, vertex_names, max_id in
	 let vertex_numbers, vertex_names, max_id = update vertex_numbers vertex_names max_id v in
	 let vertex_numbers, vertex_names, max_id = update vertex_numbers vertex_names max_id w in
	   vertex_numbers, vertex_names, max_id)
      (StringMap.empty, IntMap.empty, -1)
      lines in
  let vertex_numbers, vertex_names, max_id =
    List.fold_left
      (fun (vertex_numbers, vertex_names, max_id) (v, w) ->
	 let update vertex_numbers vertex_names max_id v =
	   if not (is_int v) && not (StringMap.mem v vertex_numbers)
	   then
	     let i = max_id + 1
	     in
	       StringMap.add v i vertex_numbers,
	       IntMap.add vertex_names i v,
	       i
	   else vertex_numbers, vertex_names, max_id in
	 let vertex_numbers, vertex_names, max_id = update vertex_numbers vertex_names max_id v in
	 let vertex_numbers, vertex_names, max_id = update vertex_numbers vertex_names max_id w in
	   vertex_numbers, vertex_names, max_id)
      (vertex_numbers, vertex_names, max_id)
      lines in
  let g = IntMap.fold (fun g i _ -> add_vertex g i) vertex_names empty in
  let g =
    List.fold_left
      (fun g (v, w) ->
	 let i = StringMap.find v vertex_numbers in
	 let j = StringMap.find w vertex_numbers in
	   if is_connected g i j
	   then invalid_arg ("multiple edge: " ^ v ^ " " ^ w)
	   else connect g i j)
      g
      lines
  in
     g, vertex_numbers, vertex_names
;;

