(* Matchings on bipartite graphs
   Copyright (C) 2008 Manuel Sorge

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(* Util *)
let intset_to_list s = IntSet.fold begin fun l i -> i :: l end s [];;

let graph_add' g i = try Graph.add_vertex g i with IntMap.Already_present -> g;;

let list_make length content = 
  let rec loop list l =
    if l = 0 then list
    else loop (content :: list) (l - 1)
  in
    loop [] length
;;

(* Matching structure *)
type matching = Graph.t;;

let matching_make = Graph.empty;;

let matching_empty = Graph.empty;;

let matching_is_empty m = m = matching_make;;

(* ATTENTION: This function can possibly create lonesome garbage vertices.
   (This however can not happen in the context used below.) *)
let toggle_edge m i j =
  let m = graph_add' (graph_add' m j) i in
    try Graph.connect m i j
    with IntSet.Already_present -> Graph.disconnect m i j 
;;

let matched m i j = try Graph.is_connected m i j with Not_found -> false;;

let single m i = try 0 = Graph.deg m i with Not_found -> true;;

let single' m d i = try d > Graph.deg m i with Not_found -> true;;

let single'' m d i = try Graph.deg m i with Not_found -> 0;;

(* General label structure *)
type 'a label = 'a IntMap.t;;

let label_make = IntMap.empty;;

let label_set l i label = IntMap.set l i label;;

let label_get l i = IntMap.get l i;;


let label_init l vertices init =
  IntSet.fold begin fun l i ->
    label_set l i (init i)
  end vertices l
;;

(* Specific label structure *)
type slabel = {
  boy: bool;
  marked: bool;
  from: int;
  visited: bool
};;

let fallback_label = {
  boy = false;
  marked = false;
  from = -1;
  visited = false;
};;

let label_get l i = try label_get l i with Not_found -> fallback_label;;

let boy l i = (label_get l i).boy;;
let girl l i = not (boy l i);;
let marked l i = (label_get l i).marked;;
let free l i = not (marked l i);;
let sboy l i = let l = label_get l i in l.marked && l.boy;;
let tgirl l i = let l = label_get l i in l.marked && (not l.boy);;
let visited l i = (label_get l i).visited;;
let mark l i j = 
    label_set l i {(label_get l i) with marked = true; from = j}
;;
let visit l i = label_set l i {(label_get l i) with visited = true};;

let init_label m w i = 
  let sboy = single' m w i in {
      boy = true;
      marked = sboy;
      from = -1;
      visited = sboy;
    }
;;


(* Main *)
(* Main: Initial matching with greedy strategy *)
let maximal_matching g =
  let m = matching_make in
    FGraph.fold_edges begin fun m i j -> 
      if (single m i) && (single m j) then toggle_edge m i j else m
    end g m
;;

let maximal_matching_bipartite g v ?(w = 1) () =
(*  if w = 1 then maximal_matching g
  else *)
    IntSet.fold begin fun m i ->
      let n = IntSet.minus (FGraph.neighbors g i) v in
	IntSet.fold begin fun m j ->
	  if (single' m w i) && (single m j) then toggle_edge m i j else m
	end n m
    end v matching_make
;;

(* Main: Augmenting paths via dfs *)
let rule1 l i j e w =
  if boy l i && free l i && tgirl l j then true, mark l i j, e
  else if boy l j && free l j && tgirl l i then true, mark l j i, e 
  else begin (* Printf.printf 
      "Rule1 failed: j: %i; i: %i; boyj: %b; freej: %b; girli: %b; freei: %b\n" 
      j i (boy l j) (free l j) (girl l i) (free l i); *) false, l, e 
  end
;;

let rule2 m l i j e =
  if sboy l i && girl l j && free l j then    
    true, mark l j i, (if single m j then j :: e else e)
  else if sboy l j && girl l i && free l i then 
    true, mark l i j, (if single m i then i :: e else e)
  else begin (* Printf.printf 
      "Rule2 failed: j: %i; i: %i; boyj: %b; freej: %b; girli: %b; freei: %b\n" 
      j i (boy l j) (free l j) (girl l i) (free l i); *) false, l, e 
  end
;;

let label g boys m w =
  let rec loop vertices (l, e) =
    match vertices with
	[] -> (l, e)
      | i :: rest ->
	  let v', l, e =
	    FGraph.fold_neighbors begin fun (v, l, e) j ->
	      let add, l, e =
		if matched m i j then rule1 l i j e w
		else rule2 m l i j e 
	      in
	      let v' = 
		if add && not (visited l j) then begin
(*		  (List.rev (j :: (List.rev v))) *)
		  j :: v
		end else v
	      in v', (if add then visit l j else l), e
	    end g i (rest, l, e)
	  in loop v' (l, e)
  in
  let l = label_init label_make boys (init_label m w) in
  let q = intset_to_list (IntSet.filter (single' m w) boys) in
    loop q (l, [])
;;

let rec construct_path startv l =
  let rec loop v path odd =
    let pre = (label_get l v).from in
    if pre = -1 then begin assert odd; path end
    else loop pre (pre :: path) (not odd)
  in
    loop startv [startv] false
;;

(* Main: Augment the matching *)
let rec augment m p =
  match p with
      [] |  _ :: [] -> m
    | i :: ((j :: _) as rest) -> augment (toggle_edge m i j) rest
;;

let remove_augmenting_paths g boys m w =
  let l, last_girls = label g boys m w in
  let fold (s, m, l) last_girl =
    let p = construct_path last_girl l in
(*      List.iter begin fun i -> Printf.printf "%i;" i end p; *)
      true, augment m p, l
  in
  let s, m, _ = List.fold_left fold (false, m, l) last_girls in
    s, m
;;

(* Main: main *)
let maximum_matching_bipartite graph v ?(w = 1) ?(m = matching_make) () =
  let m = 
    if matching_is_empty m then maximal_matching_bipartite graph v ~w () else m
  in
  let rec loop m =
    let s, m = remove_augmenting_paths graph v m w in
      if not s then m else loop m
  in 
    loop m
;;

(* Wrap *)
let maximum_edge_centered_starpacking graph v weight ?(m = matching_make) () =
  maximum_matching_bipartite graph v ~w:weight ~m:m ()

