(* solve the Maximum s-Plex problem
   Copyright (C) 2009 Hannes Moser, Manuel Sorge

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

let sort_vertices g =
  let l = Graph.fold_vertices (fun l i _ -> i :: l) g [] in
  let l = List.stable_sort (fun i j -> compare (Graph.deg g i) (Graph.deg g j)) l in
  let old_of_new, new_of_old =
    List.fold_left
      (fun (old_of_new, new_of_old) i ->
         let i' = if IntMap.is_empty old_of_new then 0 else (IntMap.max_key old_of_new) + 1 in
           IntMap.add old_of_new i' i, IntMap.add new_of_old i i')
      (IntMap.empty, IntMap.empty) l in
  let g' = IntMap.fold (fun g' i _ -> Graph.add_vertex g' i) old_of_new Graph.empty in
  let g' =
    Graph.fold_edges
      (fun g' i j ->
         Graph.connect g' (IntMap.get new_of_old i) (IntMap.get new_of_old j))
      g g'
  in
    g', old_of_new, new_of_old
;;


let die fmt =
  Printf.kprintf
    (fun s -> Printf.eprintf "%s\n" s; exit 1) fmt
;;

let usage_msg = "Find s-plexes";;

(* type isol_type = No_isol | Max_isol of int | Min_isol of int | Avg_isol of int
		 | Avg_isol_bf of int
;; *)

let verbose              = ref false;;
let dotfiles             = ref false;;
let s                    = ref 1;;
let format               = ref 1;;
let enumeration          = ref false;;
let compl                = ref false;;
let pivoting             = ref false;;
let from_below           = ref false;;
let time_limit           = ref (- 1.0);;
(* let itco                 = ref false;; *)

let setref r b =
  r := b;; 

let specs = [
  ("-v",   Arg.Set(verbose),
           "Print progress to stderr");
  ("-s",   Arg.Int(fun val_s -> if val_s < 1 then 
                                  raise (Arg.Bad "wrong argument `s'; option '-s' expects an integer > 0") 
                                else s := val_s),
           "Set s-plex parameter (default 1)");
  ("-format", Arg.Int(fun val_s -> if (val_s < 1) || (val_s > 2) then
			raise (Arg.Bad "wrong argument `format'; option '-format' expects an integer between 1 and 2") 
		      else format := val_s),
           "Specify the input format. 1: one vertex or one edge per line or 2: DIMACS ASCII format.");
  ("-ei",  Arg.Bool(setref Bdd.exhaustive_interleaving),
           "Exhaustive interleaving (default: " ^ string_of_bool !Bdd.exhaustive_interleaving ^ ")");
(*  ("-c",   Arg.Bool(setref compl),
           "Search s-plex in complementary graph. (default: " ^ string_of_bool !compl ^ ")"); *)
  ("-ii",  Arg.Int(fun val_s -> if val_s < 1 then 
                                  raise (Arg.Bad "wrong argument `ii'; option '-ii' expects an integer > 0") 
                                else Bdd.interleaving_interval := val_s),
           "Set level-interval for interleaving in the search tree. (default " ^ string_of_int !Bdd.interleaving_interval ^ ")");
  ("-tl",  Arg.Int(fun val_s -> if val_s < 1 then 
                                  raise (Arg.Bad "wrong argument `tl'; option '-tl' expects a integer > 0") 
                                else time_limit := (float_of_int val_s)),
           "Specify a time limit in seconds. (default " ^ string_of_float !time_limit ^ ")");
  ("-b",   Arg.Bool(setref Bdd.buss_interleaving),
           "High degree interleaving in search tree (default: " ^ string_of_bool !Bdd.buss_interleaving ^ ")");
  ("-d",   Arg.Bool(setref Bdd.deg_one_interleaving),
           "Degree-one interleaving in search tree (default: " ^ string_of_bool !Bdd.deg_one_interleaving ^ ")");
  ("-er",  Arg.Bool(setref Bdd.edge_rule),
           "Use the edge count to rule out branches without solutions. (Edge-count rule.) (default: " ^ string_of_bool !Bdd.edge_rule ^ ")");
  ("-br",  Arg.Bool(setref Bdd.bound_rule),
           "Calculate a lower bound to rule out branches without solutions (default: " ^ string_of_bool !Bdd.bound_rule ^ ")");
  ("-ibr", Arg.Bool(setref Bdd.improve_lower_bound),
           "Try to improve the lower bound by substituting single stars in the packing with multiple stars. (default: " ^ string_of_bool !Bdd.improve_lower_bound ^ ")");
  ("-d2",   Arg.Bool(setref Bdd.deg_two_interleaving),
           "Degree-two interleaving in search tree for s = 1 (default: " ^ string_of_bool !Bdd.deg_two_interleaving ^ ")");
  ("-dd",   Arg.Bool(setref Bdd.degree_less_than_or_equal_d_interleaving),
           "Remove vertices whose closed neighborhood has degree <= d. (default: " ^ string_of_bool !Bdd.degree_less_than_or_equal_d_interleaving ^ ")");
  ("-bdd",  Arg.Bool(setref Bdd.nt_kernelization),
           "Apply BDD-rule interleaving. (default: " ^ string_of_bool !Bdd.nt_kernelization ^ ")");
  ("-g",   Arg.Int(setref Bdd.greedyopt_interval),
           "Set to n, to optimize greedy solution in each n'th search tree level. If n <= 0 then the greedy solution is never optimized. (default: " ^ string_of_int !Bdd.greedyopt_interval ^ ")");
  ("-vs",   Arg.Bool(setref Bdd.guided_branch_vertex_selection),
           "Select branch vertex guided by greedy solution (default: " ^ string_of_bool !Bdd.guided_branch_vertex_selection ^ ")");
  ("-gb",  Arg.Bool(setref Bdd.greedysol_branching),
           "Try to branch in greedy solution (default: " ^ string_of_bool !Bdd.greedysol_branching ^ ")");
  ("-gbn",   Arg.Int(fun val_s -> if val_s < 1 then 
                                  raise (Arg.Bad "wrong argument `gbn'; option '-gbn' expects an integer > 0") 
                                else Bdd.min_branch_candidates := val_s),
           "Set minimum set of branch candidates. (default " ^ string_of_int !Bdd.min_branch_candidates ^ ")");
  ("-fb",  Arg.Bool(setref Bdd.forbid_branch_vertices),
           "Forbid branch vertices in further branching. (default: " ^ string_of_bool !Bdd.forbid_branch_vertices ^ ")");
  ("-fs",  Arg.Bool(setref Bdd.forbid_sol_subsets),
           "Forbid partial solutions of returning tree nodes. (default: " ^ string_of_bool !Bdd.forbid_sol_subsets ^ ")");
(*  ("-e",   Arg.Set(enumeration),
           "Find a maximum s-plex by exhaustive enumeration"); *)
  ("-blw",  Arg.Bool(setref from_below),
           "Increase parameter k for BDD instance, starting with lower bound (default: " ^ string_of_bool !from_below ^ ")");
(*  ("-p",   Arg.Set(pivoting),
           "Try every vertex as pivot"); *)
(*  ("-i",   Arg.Set(itco),
           "Use iterative compression"); *)
  ("-D",   Arg.Set(dotfiles),
           "Create debug dot files of intermediary graphs");
];;

let debug (s: ('a, 'b, 'c) format) = let x = Printf.eprintf (s: ('a, 'b, 'c) format) in flush stderr; x;;

let graphinfo g =
  let n = FGraph.num_vertices g in
  let m = FGraph.num_edges g in
  let f_n = float_of_int n in
  let f_m = float_of_int m in
  let density = f_m /. ((f_n *. (f_n -. 1.0)) /. 2.0) in
  debug "%i vertices, %i edges, density = %f\n" n m density
;;

let rec closed_l_neighborhood g s l = 
  if l <= 0 then IntSet.empty else
    let nb = IntSet.fold (fun accu i -> IntSet.union accu (FGraph.neighbors' g i)) s IntSet.empty in
    IntSet.union nb (closed_l_neighborhood g nb (l-1))
;;

let peel g d =
(*  let t_start = Util.timer () in *)
  let g = if not !compl then FGraph.complement g else g in
(*    Printf.eprintf "C: %f" (t_start -. (Util.timer ())); *)
  let greedy_solution = Bdd.bddsolve_greedy_f g d in
  let k = IntSet.size greedy_solution in
  let s, g, k = Bdd.bddkernel_buss g d k in
    if not !compl then FGraph.complement g else g
;;

let measure l =
  let pre = Util.timer () in
    Lazy.force l, pre -. (Util.timer ())
;;
  
let numkernelizations = ref 0;;

(* Reset some counters and timers *)
Bdd.nodecount := 0;;
Bdd.bounded_degree_nodecount := 0;;
Bdd.finished_after_simple_interleaving_nodecount := 0;;
Bdd.no_instance_after_simple_interleaving_nodecount := 0;;
Bdd.bound_rule_nodecount := 0;;
Bdd.edge_rule_nodecount := 0;;
Bdd.no_instance_due_forbidden_vertices_nodecount := 0;; 
Bdd.rrules_time := 0.0;;
Bdd.branch_vertex_selection_time := 0.0;;
Bdd.greedy_sol_time := 0.0;;
Bdd.preprocess_time := 0.0;;
Bdd.edge_rule_time := 0.0;;

let () =
  Arg.parse specs (fun _ -> Arg.usage specs usage_msg) usage_msg;

(* Input the graph. *)
  let line_parser = 
    match !format with
	1 -> Graph.line_parser_fsu
      | 2 -> Graph.line_parser_dimacs_asc
  in
  let g, vertex_numbers, vertex_names = Graph.input stdin line_parser in

(* Initialise the wrapper. *)
  let g = FGraph.of_graph g in
  let timer_start = Util.timer () in

(* Output some information about the graph. *)
(* TODO: Debug text is wrong for !compl. *)
  if !verbose then debug "Input Graph: (s-plex instance)\n    ";
  if !verbose then graphinfo g;
 
  (* Apply peeling (low-degree) reduction rule. *)
  let d = !s-1 in
  let prepeel_vertices = FGraph.num_vertices g in
  let g = peel g d in
  let peel_vertices_reduced = prepeel_vertices - (FGraph.num_vertices g) in
  if !verbose then debug "Input graph after applying low-degree reduction rule (peeling):\n    ";
  if !verbose then graphinfo g;

  let g' = if not !compl then FGraph.complement g else g in
  if !verbose then debug "Complement Graph: (bdd instance)\n    ";
  if !verbose then graphinfo g';
 
(* Calculate the bounds. *)
  let t_start = Util.timer () in 
  let greedy_sol = Bdd.bddsolve_greedy_f g' d in
  let upper_bound = IntSet.size greedy_sol in
  if !verbose then debug "Bdd-solution size upper bound is %i\n" upper_bound;
  if !verbose then debug "Time needed for computing upper bound: %f\n" (Util.timer () -. t_start);

  let t_start = Util.timer () in 
  let lower_bound = Bdd.lower_bound_f g' d in
  if !verbose then debug "Bdd-solution size lower bound is %i\n" lower_bound;
  if !verbose then debug "Time needed for computing lower bound: %f\n" (Util.timer () -. t_start);

  let timer_preprocess = Util.timer () in
  if !verbose then debug "Time elapsed so far: %f\n" (timer_preprocess -. timer_start);
  flush stderr;

(* Parameter probing loops. *)
  let rec loop g' k time_used =
    if !verbose then debug "Trying k = %i\n" k;
    flush stderr;
    
    let (continue, time_out, sol), time_used' = 
      measure 
	(lazy
	   (try 
	      false, false, Bdd.bddsolve' g' d k (!time_limit -. time_used)
	    with 
		Bdd.NO_instance -> true, false, IntSet.empty
	      | Bdd.Time_out -> false, true, IntSet.empty))
    in
      if time_out then begin
	if !verbose then debug "Time out.";
	IntSet.empty, k - 1, upper_bound
      end else if k >= upper_bound && continue then 
	die "Error: bddsolve did not find a solution for an instance with k >= greedy upper bound. This should never happen."
      else if k >= upper_bound || (not continue) then 
	sol, IntSet.size sol, IntSet.size sol
      else loop g' (k + 1) (time_used +. time_used')
  in
    
  let rec loop' g' k sol time_used =
    if k >= lower_bound then begin
      if !verbose then debug "Trying k = %i\n" k;
      flush stderr;
      let (continue, time_out, sol), time_used' =
	measure
	  (lazy
	     (try 
		true, false, Bdd.bddsolve' g' d k (!time_limit -. time_used)
	      with
		  Bdd.NO_instance -> false, false, sol
		| Bdd.Time_out -> false, true, sol))
      in
	if not continue then
	  if time_out then begin
	    if !verbose then debug "Time out.\n";
	    sol, lower_bound, k + 1
	  end else
	    sol, IntSet.size sol, IntSet.size sol
	else
	  loop' g' ((IntSet.size sol) - 1) sol (time_used +. time_used')
    end else sol, IntSet.size sol, IntSet.size sol
  in
    
(* Call-function for probing. *)
  let rec solve g time_used =
    (* TODO: add timer for the following/cut it intelligently (redundant bound computing) *)
    let g' = if not !compl then FGraph.complement g else g in
    let greedy_sol = Bdd.bddsolve_greedy_f g' d in
    let upper_bound = IntSet.size greedy_sol in
    let lower_bound = Bdd.lower_bound_f g' d in 
    if upper_bound = lower_bound then 
      greedy_sol, IntSet.size greedy_sol, IntSet.size greedy_sol
    else (* if !enumeration then 
      Bdd.bddsolve g' d upper_bound
    else *) if !from_below then
      loop g' lower_bound time_used
    else
      loop' g' (upper_bound - 1) greedy_sol time_used in
(* TODO: Pivoting with time limit. 
  let rec solve_pivot g =
    let vertices = FGraph.vertices g in
    let v = IntSet.choose vertices in
    let nb = closed_l_neighborhood g (IntSet.add IntSet.empty v) !s in
    let remaining_vertices = IntSet.minus vertices nb in
    if IntSet.size remaining_vertices < 20 then
       solve g
    else
      let gg = FGraph.subgraph g nb in 
      let sol1 = solve gg in
      let sol2 = solve_pivot (FGraph.delete_vertex g v) in
      if IntSet.size sol1 + IntSet.size remaining_vertices < IntSet.size sol2 + 1 then
        IntSet.union sol1 remaining_vertices 
      else IntSet.add sol2 v
  in  
*)
  Bdd.nodecount := 0;

(*  if !itco && (!s > 2 || !pivoting) then
    begin
      debug "Iterative compression is only possible for s <= 2 without pivoting\n";
      exit 1 
    end
  else *)
  let time_used = timer_preprocess -. timer_start in
  let sol, best_lower_bound, best_upper_bound =(* if !pivoting then 
              solve_pivot g time_used
            else*) (* if !itco then 
              begin
                debug "WARNING: Iterative compression is not yet fully implemented!\n";
                Bdd.bddsolve_itco (if not !compl then FGraph.complement g else g) d
              end
            else *) solve g time_used in

  let timer_end = Util.timer () in

(* Output the gathered debug information in Bdd. *)
  if !verbose && best_lower_bound == best_upper_bound then debug "Maximum s-plex size: %i\n" ((FGraph.num_vertices g) - (IntSet.size sol));
  if !verbose && best_lower_bound <> best_upper_bound then debug "Maximum s-plex size between %i and %i.\n" ((FGraph.num_vertices g) - best_upper_bound) ((FGraph.num_vertices g) - best_lower_bound);
  if !verbose then debug "Number of applications of high-degree rule, number of reduced vertices: %i, %i\n" !Bdd.buss_num (!Bdd.buss_vertices_deleted + peel_vertices_reduced);
  if !verbose then debug "Number of applications of degree one rule, number of reducted vertices: %i, %i\n" !Bdd.deg_one_num !Bdd.deg_one_vertices_deleted;  
  if !verbose then debug "Number of applications of degree two rule: %i\n" !Bdd.deg_two_num;
  if !verbose then debug "Number of applications of clique rule, number of reducted vertices: %i, %i\n" !Bdd.cliquerule_num !Bdd.cliquerule_vertices_deleted;  
  if !verbose then debug "Number of applications of BDD-rule, number of reduced vertices: %i, %i\n" !Bdd.ntrule_num !Bdd.ntrule_vertices_deleted;
  if !verbose then debug "Number of search tree nodes: %i\n" !Bdd.nodecount;
  if !verbose then debug "Number of search tree nodes with bounded-degree input: %i\n" !Bdd.bounded_degree_nodecount;
  if !verbose then debug "Number of search tree nodes with no-instance input due to simple interleaving: %i\n" !Bdd.no_instance_after_simple_interleaving_nodecount;
  if !verbose then debug "Number of search tree nodes with yes-instance input due to simple interleaving: %i\n" !Bdd.finished_after_simple_interleaving_nodecount; 
  if !verbose then debug "Number of search tree nodes rejected due to bound-rule: %i\n" !Bdd.bound_rule_nodecount;
  if !verbose then debug "Number of search tree nodes rejected due to edge rule: %i\n" !Bdd.edge_rule_nodecount;
  if !verbose then debug "Number of search tree nodes rejected due to forbidden vertices: %i\n" !Bdd.no_instance_due_forbidden_vertices_nodecount;
  if !verbose then debug "Number of search tree nodes rejected due to branching: %i\n" !Bdd.no_instance_due_branching_nodecount;
  if !verbose then debug "Number of search tree nodes rejected due to forbidden solution subset: %i\n" !Bdd.forbidden_nodecount;
  if !verbose then debug "Total time spent in simple data reduction rules: %f\n" !Bdd.rrules_time;
  if !verbose then debug "Total time spent for selecting the branch vertex: %f\n" !Bdd.branch_vertex_selection_time;
  if !verbose then debug "Total time spent for optimizing the greedy solution: %f\n" !Bdd.greedy_sol_time;
  if !verbose then debug "Total time spent for edge rule: %f\n" !Bdd.edge_rule_time;
  if !verbose then debug "Total time spent for lower bound rule: %f\n" !Bdd.lb_rule_time;
  if !verbose then debug "Total time spent in search tree nodes before branching: %f\n" !Bdd.preprocess_time;
  if !verbose then debug "Number of executions of lower bound rule: %i\n" !Bdd.lb_executions;
  if !verbose then debug "Number of improvements of lower bound: %i\n" !Bdd.lb_improvements;

(* Output the maximum s-plex. *)
   Printf.printf "%s\n" 
    (IntSet.fold 
       (fun s i -> Printf.sprintf "%s %s" vertex_names.(i) s) 
       (IntSet.minus (FGraph.vertices g) sol)
       "");


  (*Graph.iter_edges (fun i j -> Printf.eprintf "%s %s\n" vertex_names.(i) vertex_names.(j)) (Graph.complement (FGraph.to_graph (FGraph.subgraph g (IntSet.minus (FGraph.vertices g) sol))));
  *)

(* The last line of debug output contains some summary information
   about the input graph and the execution of the solver. The following
   list describes the meaning of each value, ordered by column number.

   1     Number of input-vertices
   2     Number of input-edges
   3     Number of edges in the complement-graph
   4     Bdd solution size
   5     Maximum s-plex size
   6     Node count
   7,8,9 Overall time needed, time needed for solve, preprocessing time
   10    Number of reduced vertices per search tree node for high-degree rule
   11    Number of reduced vertices per search tree node for degree one rule
   12    Number of applications per search tree node of the degree two rule (s = 1)
   13,14 Number of applications and number of reduced vertices per search tree node for BDD-rule 
   15,16 Best bounds for s-plex size
*)
  let int_int_divider (i:int) (j:int) =  (float_of_int i) /. (float_of_int j) in
  let node_divider i = int_int_divider i !Bdd.nodecount in

  if !verbose then Printf.printf "%d\t%d\t%d\t%d\t%d\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%i\t%i\n" 
    (FGraph.num_vertices g) 
    (FGraph.num_edges g) 
    (FGraph.num_edges g') 
    (IntSet.size sol) 
    ((FGraph.num_vertices g) - (IntSet.size sol)) 
    !Bdd.nodecount 
    (timer_end -. timer_start) (timer_end -. timer_preprocess) (timer_preprocess -. timer_start)
    (node_divider (!Bdd.buss_vertices_deleted + peel_vertices_reduced))
    (node_divider !Bdd.deg_one_vertices_deleted)
    (node_divider !Bdd.deg_two_num)
    (node_divider !Bdd.ntrule_num) (node_divider !Bdd.ntrule_vertices_deleted)
    ((FGraph.num_vertices g) - best_upper_bound)
    ((FGraph.num_vertices g) - best_lower_bound)
    ;
()
;;
